/*
 * Decode NEC IR RC protocol..
 */

#define TIMEOUT  150000
#define PULSES_MAX  34
#define TOLERANCE  15  /* % */

// TSOP1736 connected to pins 8 (gnd), 9 (Vcc) & 11 (signal)
int ir_signal = 11;
int ir_vcc = 9;
int ir_gnd = 8;

typedef struct
{
  unsigned long timestamp;
  unsigned long on_time;
}
point_t;

point_t pulses[PULSES_MAX];
int index;

// Times in microseconds
uint16_t startbit = 9000;
uint16_t on_time_0 = 560;
uint16_t off_time_0 = 565;
uint16_t on_time_1 = 560;
uint16_t off_time_1 = 1690;
unsigned long time_start;
unsigned long rc_code = 0;

boolean interval_within_bounds(uint16_t interval, uint16_t target)
{
  unsigned long dx = (unsigned long)target*TOLERANCE/100;
  return (interval>target-dx && interval<target+dx);
}

boolean bit_add(uint8_t value, uint16_t interval, uint16_t target)
{
  if (interval_within_bounds(interval,target)==true)
  {
    rc_code <<= 1;
    if (value!=0) rc_code |= 1;
    return true;
  }
  return false;
}

boolean decode_nec(void)
{
  int i;
  uint16_t interval;
  unsigned long t_prev = pulses[0].timestamp;
  uint8_t pulsecount = PULSES_MAX;
  
  rc_code = 0;
  for (i=1; i<pulsecount; i++)
  {
    // Four intervals are possible: normal "0" and "1" bits, 1st data bit & repeat code.
    interval = pulses[i].timestamp - t_prev;
    if (interval_within_bounds(interval,on_time_0+startbit/2)==true)
    {
      // First data bit.
      if (interval_within_bounds(pulses[i].on_time,on_time_0)==false)
      {
        goto decode_nec_error;
      }
    }
    else if (interval_within_bounds(interval,on_time_0+startbit/4)==true)
    {
      // Repeat code.
      if (interval_within_bounds(pulses[i].on_time,on_time_0)==false)
      {
        goto decode_nec_error;
      }
      pulsecount = 2;
    }
    else if (interval_within_bounds(interval,on_time_0+off_time_0)==true)
    {
      // Previous bit was a "0".
      if (bit_add(0,pulses[i].on_time,on_time_0)==false)
      {
        goto decode_nec_error;
      }
    }
    else if (interval_within_bounds(interval,on_time_1+off_time_1)==true)
    {
      // Previous bit was a "1".
      if (bit_add(1,pulses[i].on_time,on_time_1)==false)
      {
        goto decode_nec_error;
      }
    }
    else
    {
        goto decode_nec_error;
    }
    t_prev = pulses[i].timestamp;
  }
  return true;
  
decode_nec_error:
  Serial.print("bit ");
  Serial.print(i);
  Serial.print(", interval=");
  Serial.print(interval);
  Serial.print(", pulse=");
  Serial.print(pulses[i].on_time);
  Serial.print(", rc_code=");
  Serial.println(rc_code,HEX);
  return false;
}

boolean decode(boolean startbit_only)
{
  if (interval_within_bounds(pulses[0].on_time,startbit)==true)
  {
    if (startbit_only==true)
    {
      return true;
    }
    else
    {
      return decode_nec();
    }
  }
  return false;
}

void setup(void)
{
  Serial.begin(115200);
  pinMode(ir_signal,INPUT);
  pinMode(ir_gnd,OUTPUT);
  digitalWrite(ir_gnd,LOW);
  pinMode(ir_vcc,OUTPUT);
  digitalWrite(ir_vcc,HIGH);
  // Ignore spurious pulses at startup.
  delay(500);
  while (pulseIn(ir_signal,LOW,500000)!=0);
}

void loop(void)
{
  unsigned long pulse = pulseIn(ir_signal,LOW,TIMEOUT);
  if (pulse!=0)
  {
    unsigned long time_now = micros();
    if (time_start==0)
    {
      time_start = time_now;
    }
    if (index<=PULSES_MAX)
    {
      pulses[index].timestamp = time_now - time_start;
      pulses[index].on_time = pulse;
      if (index==0 && decode(true)==false)
      {
      	// resynch
        time_start = 0;
      }
      else index += 1;
    }
  }
  else
  {
    if (index>0)
    {
      if (decode(false)==true)
      {
        Serial.println(rc_code,HEX);
      }
      index = 0;
    }
    time_start = 0;
  }
}
